@@ -1,6 +1,4 @@
-from torchcam import cams
-from torchcam import utils
-
+from torchcam import cams, utils
 
 try:
     from .version import __version__  # noqa: F401
