@@ -1,5 +1,3 @@
 from .cam import *
 from .gradcam import *
-
-del cam
-del gradcam
+from .utils import *
